#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/mman.h>
#ifndef __APPLE__
#include <sys/sendfile.h>
#else
#define RLIMIT_OFILE RLIMIT_NOFILE
#define SENDFILE
#endif
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <poll.h>
#include "mImpl.h"

#include <vector>
#include <map>
#include <deque>
#include <list>

#define ERR_READ  -1
#define ERR_WRITE -2

typedef enum
  {
    S_STATE_FREE,
    S_STATE_CONNECTED,
    S_STATE_ACTIVE,
    S_STATE_COUNT
  } socket_state_t;

enum
{
	REQMODE_NORMAL,
	REQMODE_FULLREAD,
	REQMODE_MMAP,
	REQMODE_SENDFILE,
	REQMODE_MAX
};

char * reqmode_table[] = {
	"normal",
	"fullread",
	"mmap",
	"sendfile",
	NULL
};

socket_state_t *socket_map = NULL;

int socket_map_size;
int max_socket;

int reportIOTime = 0;
int requestMode = REQMODE_FULLREAD;

int listen_socket;
struct sockaddr_in server_addr;
char *root;
int root_len;

fd_set read_fds;
int fd_max;

int suffixTest(char *val, char *suffix) {
  int len = strlen(val);
  int s_len = strlen(suffix);
  int i;
  
  for (i=0;i<s_len;i++) {
    if (val[len-i]!=suffix[s_len-i])
      return 0;
  }
  return 1;
}


void returnSocket(int sock) 
{
  socket_map[sock] = S_STATE_CONNECTED;
}

void closeSocket(int sock) {
  if (sock > -1) 
    {
      close(sock);
      socket_map[sock] = S_STATE_FREE;
    }
}



void init(int argc, char **argv) {
  signal(SIGPIPE, SIG_IGN);

  int old_flags;
  int offset = 0;

  struct rlimit r;
  getrlimit(RLIMIT_OFILE,&r);
  socket_map_size = r.rlim_max;
  if (socket_map_size < 0) {
	perror("Get max sockets");
	fprintf(stderr, "Can not detect maximum # of sockets, defaulting to 10\n");
	socket_map_size = 10;
  }

  printf("Max socket: %d\n", socket_map_size-1);
  socket_map = new socket_state_t[socket_map_size];
  bzero(socket_map,sizeof(socket_state_t) * socket_map_size);
      
  listen_socket = socket(AF_INET,SOCK_STREAM,0);
  max_socket = listen_socket;

  int val = 1;
    if (setsockopt(listen_socket, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val)) < 0)
    {
        perror("setsockopt: ");
        exit(1);
    }

	while(argc > 1)
	{
		if (!strcasecmp(argv[offset + 1], "--reportIOTime"))
		{
			reportIOTime=1;
			offset++;
			argc--;
			continue;
		}

		if (!strncasecmp(argv[offset + 1], "--requestMode=", strlen("--requestMode=")))
		{
			int i = 0;

			for (i = 0; i < REQMODE_MAX; i++)
				if (!strcasecmp(reqmode_table[i], argv[offset + 1] + strlen("--requestMode=")))
					break;

			if (i >= REQMODE_MAX)
			{
				printf("Invalid request mode.  Valid modes:\n");

				for (i = 0; i < REQMODE_MAX; i++)
					printf("%s\n", reqmode_table[i]);

				exit(1);
			}

			requestMode = i;

			offset++;
			argc--;
			continue;
		}

		break;
	}
 
  if (argc < 3) {
     fprintf (stderr, "Usage: %s [--reportIOTime] [--requestMode=mode] <port-number> <root-dir>\n", argv[0]);
     exit(1);
  }

  server_addr.sin_family = AF_INET;
  server_addr.sin_port = htons(atoi(argv[offset+1]));
  server_addr.sin_addr.s_addr = htonl(INADDR_ANY);

  root = argv[offset+2];
  root_len = strlen(root);

  printf("setting root directory to [%s]\n", root);
  if (chdir(root) < 0)
    {
      perror("chdir");
      exit(1);
    }
  
  if ((bind(listen_socket,(struct sockaddr*) &server_addr, sizeof(struct sockaddr))) < 0) {
    perror("Bind");
    exit(-1);
  }
  if (listen(listen_socket,50000)<0) {
    perror("Listen");
    return;
  }

  if (server_addr.sin_addr.s_addr) {
    fprintf (stderr, "listening on address %s, port %d\n",
	     server_addr.sin_addr.s_addr, atoi(argv[offset+1]));
  } else {
    fprintf (stderr, "listening on address (NULL), port %d\n",
	     atoi(argv[offset+1]));
  }

	printf("Using %s request mode\n", reqmode_table[requestMode]);
}


int Reply (Reply_in *in) {
  if (in->close) 
    closeSocket(in->socket);
  else 
    returnSocket(in->socket);
  return 0;
}

#define BUFFER_SIZE 8192

int ReadRequest (ReadRequest_in *in, ReadRequest_out *out) {
  char buf[BUFFER_SIZE];
  int rd;
  int length = 0; 
  int doneRequest = 0;
  out->request = 0;
  out->close = false;

  //DEBUG printf("ReadRequest in\n");
  do {
    rd = read(in->socket, buf+length, BUFFER_SIZE-length);
    if (rd == -1) {
      perror("Reading request");
      fprintf(stderr,"socket was %d\n\n", in->socket);
      closeSocket(in->socket);
      return -1;
    } 
    else if (!rd) {
      usleep(1000*10);
    }
    else {
      char *start = buf;
      char *end;
      for (int i=0;i<rd;i++) {
	if (buf[i] == '\r') {
	  buf[i++] = 0;
	  buf[i] = 0; // Get rid of the \n
	  //DEBUG printf("%s\n", start);
	  if (length == 0) { // We're done...
	    doneRequest = true;
	    break;
	  }
	  else if (start[0] == 'G' && start[1] == 'E' && start[2] == 'T') {
	    start = strchr(start, ' ')+1;
	    end = strchr(start+1, ' ');
	    *end = 0;
	    while (*end != 'H')
	      end++;
	    while (*end != '/')
	      end++;
	    end++;
	    int major = *end-'0'; // HACK HACK HACK Assumes ASCII
	    end+=2;
	    int minor = *end-'0'; // HACK HACK HACK Assumes ASCII
	    if (major != 1 || (minor > 1))
	      printf("Urm, HTTP version: %d.%d we may be in trouble...\n", 
		     major, minor);
	    if (major == 1 && minor == 0) {
	      out->close = true;
	    }
	    
	    if (*start == '/')
	      start++;
	    out->request = strdup(start);
	  }
	  else if (start[0] == 'C' && start[1] == 'o' && start[2] == 'n' &&
		   strstr(start, "close")) {
	    out->close = true;
	  }
	  start=buf+i+1;
	  length = 0;
	}
	else {
	  length++;
	}
      }
      if (!doneRequest) {
	strncpy(buf, start, length);
      }
    }  
  } while (!doneRequest);

  if (!out->request) {
    out->request = "/sys_error.html";
  }

  out->socket = in->socket;
  // DEBUG printf("Request:%s:\n", out->request);
  return 0;
}

long unsigned int get_elapsed_time(struct timeval atime, struct timeval btime)
{
    struct timeval diff;

    diff.tv_sec = btime.tv_sec - atime.tv_sec;

    if (btime.tv_usec < atime.tv_usec)
    {
        diff.tv_usec = (btime.tv_usec + 1000000) - atime.tv_usec;
        diff.tv_sec--;
    }
    else
        diff.tv_usec = btime.tv_usec - atime.tv_usec;

    return (diff.tv_sec * 1000000) + diff.tv_usec;
}

int get_count(char * file_name)
{
	struct file_entry
	{
		char*	filename;
		int		count;
		struct file_entry * next;
	};

	static struct file_entry * file_list = NULL;
	static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	struct file_entry * f;
	int i;

	pthread_mutex_lock(&mutex);

	for (f = file_list; f != NULL; f = f->next)
		if (!strcmp(f->filename, file_name))
			break;

	if (f != NULL)
	{
		i = f->count++;
		pthread_mutex_unlock(&mutex);

		return i;
	}

	f = (struct file_entry*)malloc(sizeof(struct file_entry));
	f->count = 0;
	f->filename = strdup(file_name);

	f->next = file_list;
	file_list = f;

	i = f->count++;

	pthread_mutex_unlock(&mutex);

	return i;
}

int ReadWrite (ReadWrite_in *in, ReadWrite_out *out) {
  FILE *f;
  int rd;
  int res;
  char file_name[128];
  //char *file_name;
  long unsigned int total_time = 0;
  struct timeval start_time, end_time;
	       
  out->socket = in->socket;
  out->close = in->close;
  //sprintf(file_name, "%s/%s", root, in->file);

  // changed to match open() calls in knot exactly
  sprintf(file_name, "./%s", in->file);
  //printf("Looking for:%s:\n", file_name);

  int fd = open(file_name, O_RDONLY|O_NONBLOCK);
  if (fd < 0) {
    perror("Open");
    closeSocket(in->socket);
    return -1;
  }

  struct stat sb;
  res = fstat(fd, &sb);
  int file_size = sb.st_size;
  
  if (res < 0) {
    printf("File not found (%s)\n", file_name);
    return 404;
  }
  else {
    int ix=0;
    char *ptr;
    int header_len;
    /*
    if (suffixTest(in->file, ".html")) {
      out->content = "text/html";
    }
    else if (suffixTest(in->file, ".png")) {
      out->content = "image/png";
    }
    else if (suffixTest(in->file, ".jpg") || suffixTest(in->file, ".jpeg")) {
      out->content = "image/jpeg";
    }
    else if (suffixTest(in->file, ".gif")) {
      out->content = "image/gif";
    }
    else {
      out->content = "text/plain";
    }
    */
    out->content="text/plain";
    char hdrs[8192];
    if (in->close) {
      sprintf(hdrs, "HTTP/1.1 200 OK\r\nContent-type: %s\r\nServer: Markov 0.1\r\nConnection: close\r\nContent-Length: %d\r\n\r\n", 
	      out->content, file_size);
    }
    else {
      sprintf(hdrs, "HTTP/1.1 200 OK\r\nContent-type: %s\r\nServer: Markov 0.1\r\nContent-Length: %d\r\n\r\n", 
	      out->content, file_size);
    }
    //printf(hdrs);
    header_len = strlen(hdrs);
    int written = write(in->socket, hdrs, header_len);
    while (written < header_len) {
      if (written < 0) {
	perror("Writing");
	closeSocket(in->socket);
	return -1;
      }
      written += write(in->socket, hdrs+written, header_len-written);
    }
    //out->length = sb.st_size+header_len;

    if (reportIOTime)
      gettimeofday(&start_time, NULL);
    
    if (reportIOTime)
    {
      gettimeofday(&end_time, NULL);

      total_time += get_elapsed_time(start_time, end_time);
    }

    //out->output = (char *)malloc(out->length);
    //memcpy(out->output, hdrs, header_len);
    //ptr = out->output+strlen(hdrs);

	switch (requestMode)
	{
		case REQMODE_NORMAL:
		{
		    char *pos;
			do
			{
				if (reportIOTime)
					gettimeofday(&start_time, NULL);
      
			    if ((rd = read(fd, hdrs, 8192)) < 0)
				{
					perror("read");
					closeSocket(in->socket);
					return 0;
			    }

			      if (reportIOTime)
			      {
			        gettimeofday(&end_time, NULL);

			        total_time += get_elapsed_time(start_time, end_time);
			      }

			      ix += rd;
			      pos = hdrs;

			    while (rd > 0)
				{
					int ret = write(in->socket, pos, rd);

					if (ret < 0)
					{
					  perror("write");
					  closeSocket(in->socket);
					  close(fd);
					  return 0;
					}
					else if (ret == 0) // Why is this here?
					{
					  rd = 0;
					  break;
					}
					else {
					  pos += ret;
					  rd -= ret;
					}
				}
			} while (ix < sb.st_size);

			break;
		}

		case REQMODE_FULLREAD:
		{
			char * buf, * pos;
			int toread = sb.st_size;
			int i = 0;

			buf = (char*)malloc(sb.st_size);

			while (toread > 0)
			{
				rd = read(fd, buf + i, toread);

				if (rd == -1)
				{
					perror("Fullread-read");
					closeSocket(in->socket);
					close(fd);
					free(buf);
					return 0;
				}

				i += rd;
				toread -= rd;
			}

			pos = buf;
			rd = sb.st_size;

			while (rd > 0)
			{
				int ret = write(in->socket, pos, rd);

				if (ret < 0)
				{
					perror("Fullread-write");
					closeSocket(in->socket);
					close(fd);
					free(buf);
					return 0;
				}
				else if (ret == 0) // Why is this here?
				{
					rd = 0;
					break;
				}
				else
				{
					pos += ret;
					rd -= ret;
				}
			}

			free(buf);
			break;
		}

		case REQMODE_MMAP:
		{
			char * buf, * pos;
			int i = 0;
			volatile int x = 0;

			buf = (char*)mmap(NULL, sb.st_size, PROT_READ, MAP_SHARED, fd, 0);

			if (buf == MAP_FAILED)
			{
				perror("Mmap");
				closeSocket(in->socket);
				close(fd);
				return 0;
			}

			for (rd = 0; rd < sb.st_size; rd += 1024)
                x += buf[rd];

			rd = sb.st_size;
			pos = buf;

			while (rd > 0)
			{
				int ret = write(in->socket, pos, rd);

				if (ret < 0)
				{
					perror("Mmap-write");
					closeSocket(in->socket);
					munmap(buf, sb.st_size);
					close(fd);
					return 0;
				}
				else if (ret == 0) // Why is this here?
				{
					rd = 0;
					break;
				}
				else
				{
					pos += ret;
					rd -= ret;
				}
			}

			munmap(buf, sb.st_size);

			break;
		}

		case REQMODE_SENDFILE:
		{
			int tosend = sb.st_size;
			off_t sent = 0;

			while (tosend > 0)
			{
#ifdef __APPLE__
				fprintf(stderr, "Error, SENDFILE mode currently unsupported...\n");
				rd = -1;
#else
				rd = sendfile(in->socket, fd, &sent, tosend);
#endif
				if (rd == -1)
				{
					perror("Sendfile");
					closeSocket(in->socket);
					close(fd);
					return 0;
				}

				tosend -= rd;
			}

			break;
		}
	}

    close(fd);
  }

  if (reportIOTime)
  {
     int c = get_count(file_name);
     printf("[%4d] %s: %10d bytes took %10ldusec\n", c, file_name, file_size, total_time);
  }

  //DEBUG printf("Done reading for:%s:\n", file_name);
  return 0;
}

//static std::deque <Listen_out *> *listen_outs;
static std::deque <Listen_out *> listen_outs;
static pollfd *poll_fds = NULL;
static int poll_max = 0;

int Listen(Listen_out *out)
{
  //if (listen_outs == NULL)
  //  {
  //    printf("creating new deque...\n");
  //    listen_outs = new std::deque<Listen_out *>;
  //  }

  Listen_out *var;
  if (listen_outs.empty())
    {
      if (poll_max < max_socket)
	{
	  if(poll_fds != NULL)
	    delete [] poll_fds;
	  
	  poll_fds = new pollfd[max_socket+1];
	  poll_max = max_socket;
	}
      
      poll_fds[0].fd = listen_socket;
      poll_fds[0].events = POLLIN;// | POLLOUT;
      int poll_count = 1;
      
      for(unsigned int sk = 0; sk <= max_socket; sk++)
	if(socket_map[sk] == S_STATE_CONNECTED)
	  {
	    poll_fds[poll_count].fd = sk;
	    poll_fds[poll_count].events = POLLIN;
	    poll_count++;
	  }
      
      int poll_ret = poll(poll_fds, poll_count, 500);
      if(poll_ret < 0)
	{
	  perror("Poll");
	  exit(-1);
	}
      else if (poll_ret == 0)
	{
	  //fprintf(stderr,"Poll timed out\n");
	  return -1;
	}
      
      int sock;
      for(int i = 0; i < poll_count; i++)
	{
	  if(poll_fds[i].revents & POLLIN)
	    {
	      sock = poll_fds[i].fd;
	      if(sock == listen_socket)
		{
		  socklen_t length =  sizeof(struct sockaddr);
		  sock = accept(listen_socket, (struct sockaddr *)&server_addr,&length);
		  if(sock < 0)
		    {
		      perror("Accept");
		      return -1;
		    }
		  if(sock > max_socket)
		    max_socket = sock;
		  int optval = 1;
		  if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &optval, sizeof (optval)) < 0)
		    {
		      perror("setsockopt");
		      fprintf(stderr,"socket was %d\n", sock);
		      return -1;
		    }
		}
	      var = new Listen_out();
	      var->socket = sock;
	      listen_outs.push_back(var);
	      socket_map[sock] = S_STATE_ACTIVE;
	    }
	}
    }

  if(listen_outs.empty())
    return -1;

  out->socket = (listen_outs.front())->socket;
  delete listen_outs.front();
  listen_outs.pop_front();
  
  //printf("Listen is giving back socket %d\n", out->socket);
  //printf("There are %d other things in the queue\n", listen_outs.size());
  
  return 0;
}

void writeHeaders(int socket_in, bool close, int length, char *content) {
  char msg[128];
  
  sprintf(msg, "Content-Length: %d\r\n", length);
  write(socket_in, msg, strlen(msg));
  sprintf(msg, "Server: Markov 0.1\r\n");
  write(socket_in, msg, strlen(msg));
  sprintf(msg, "Content-Type: %s\r\n", content);
  write(socket_in, msg, strlen(msg));
  if (close)
    write(socket_in, "Connection: close\r\n",19); 
  write(socket_in, "\r\n", 2);
}

void FourOhFor(ReadWrite_in *in, int err) {
  char *msg = "HTTP/1.1 404 File not found\r\n";
  write(in->socket, msg, strlen(msg));
  msg = "<html><body><h2>404 File Not Found!</h2></body></html>\n";
  writeHeaders(in->socket, in->close, strlen(msg), "text/html");
  write(in->socket, msg, strlen(msg));
  if (in->close)
    closeSocket(in->socket);
  else
    returnSocket(in->socket);
}

int SessionId(Page_in *in) {
  return in->socket;
}

void BadRequest(ReadRequest_in *in, int err) {
  char *msg;

  switch (err) {

  case 400:
    msg = "HTTP/1.1 400 Bad Request\r\n";
    if (write(in->socket, msg, strlen(msg)) != -1) {
      msg = "<html><body><h2>400 Bad Request!</h2></body></html>\n";
      writeHeaders(in->socket, true, strlen(msg), "text/html");
        write(in->socket, msg, strlen(msg));
    }
    break;

  case 501:
    msg = "HTTP/1.1 501 Not Implemted\r\n";
    if (write(in->socket, msg, strlen(msg)) != -1) {
      msg = "<html><body><h2>501 Not Implemented!</h2></body></html>\n";
      writeHeaders(in->socket, true, strlen(msg), "text/html");
        write(in->socket, msg, strlen(msg));
    }
    break;

  case 408:
    msg = "HTTP/1.1 408 Request Timeout\r\n";
    if (write(in->socket, msg, strlen(msg)) != -1) {
      msg = "<html><body><h2>408 Request Timeout!</h2></body></html>\n";
      writeHeaders(in->socket, true, strlen(msg), "text/html");
        write(in->socket, msg, strlen(msg));
    }
    break;
  }

  closeSocket(in->socket);

}

