package edu.umass.cs.flux;

import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;
import java.util.HashMap;
import java.util.Map;

/**
 * The top level representation of a Markov program
 * @author Brendan Burns
 * @version $version: $
 **/
public class Program {
    private Vector<Source> sources;
    private Vector<FlowStatement> exps;
    private Vector<ErrorHandler> errors;
    private Hashtable<String, TaskDeclaration> decls;
    private Hashtable<String, TypeDeclaration> type_decs;
    private Hashtable<String, AtomicDeclaration> atomic_defs;
	    
	    
    /**
     * Constructor
     * @param sources A Vector of Sources defining the starting events
     * @param types A Vector of TypeDeclarations defining types.
     * @param decl_list A Vector of TaskDeclarations defining functions
     * @param program A Vector of FlowStatements defining the program
     * @param atomic_decls A Vector of concurrency contraints
     * @param errors A Vector of error handlers
     **/
    public Program(Vector<Source> sources,
		   Vector<TypeDeclaration> types,
		   Vector<TaskDeclaration> decl_list,
		   Vector<FlowStatement> program,
		   Vector<AtomicDeclaration> atomic_decls,
		   Vector<ErrorHandler> errors) {
        this.sources = sources;
        this.exps = program;
	this.errors = errors;
	
	this.decls = new Hashtable<String, TaskDeclaration>();
        for (int i=0;i<decl_list.size();i++) {
            this.decls.put((decl_list.get(i)).getName(),
            				decl_list.get(i));
	}
        this.type_decs = new Hashtable<String, TypeDeclaration>();
        for (int i=0;i<types.size();i++) {
            this.type_decs.put((types.get(i)).getName(),
                    			types.get(i));
        }
	this.atomic_defs = new Hashtable<String, AtomicDeclaration>();
	for (AtomicDeclaration ad : atomic_decls) {
	    this.atomic_defs.put(ad.getName(), ad);
	}
    }

    public Source getSessionStart(FlowStatement fs) {
	for (Source s : sources) {
	    if (s.getTarget().equals(fs.getAssignee()))
		return s;
	}
	return null;
    }



    public Program(Vector<Object> line_list)
	{
		sources = new Vector<Source>();
		exps = new Vector<FlowStatement>();
		errors = new Vector<ErrorHandler>();
		decls = new Hashtable<String, TaskDeclaration>();
		type_decs = new Hashtable<String, TypeDeclaration>();
		atomic_defs = new Hashtable<String, AtomicDeclaration>();

		for (int i = 0; i < line_list.size(); i++)
		{
			Object o = line_list.get(i);

			if (o instanceof Source)
			{
				//System.out.println("Source");
				sources.add((Source)o);
			}
			else if (o instanceof FlowStatement)
			{
				//System.out.println("FlowStmt");
				exps.add((FlowStatement)o);
			}
			else if (o instanceof ErrorHandler)
			{
				//System.out.println("ErrorHandler");
				errors.add((ErrorHandler)o);
			}
			else if (o instanceof TaskDeclaration)
			{
				//System.out.println("TaskDecl");
				decls.put(((TaskDeclaration)o).getName(),(TaskDeclaration)o);
			}
			else if (o instanceof TypeDeclaration)
			{
				//System.out.println("TypeDecl");
				type_decs.put(((TypeDeclaration)o).getName(),(TypeDeclaration)o);
			}
			else if (o instanceof AtomicDeclaration)
			{
				//System.out.println("AtomicDecl");
				atomic_defs.put(((AtomicDeclaration)o).getName(),(AtomicDeclaration)o);
			}
			else
			{
				System.err.println("Unrecognized line type passed to program constructor.");
				System.exit(1);
			}
		}

    }
	
    /**
     * Get the set of atomic declarations
     * @return a Vector of AtomicDeclarations
     **/
    public Collection<AtomicDeclaration> getAtomicDeclarations() {
	return atomic_defs.values();
    }

    /**
     * Get a particular atomic declaration
     * @param name The name of the fn
     * @return The named atomic declaration 
     **/
    public AtomicDeclaration getAtomicDeclaration(String name) {
	return atomic_defs.get(name);
    }

    /**
     * Get the set of error handlers
     * @return A Vector of ErrorHandlers
     * @see ErrorHandler
     **/
    public Vector<ErrorHandler> getErrorHandlers() {
	return errors;
    }

    /**
     * Get the set of sources for this Markov program
     * @return A vector of the String names of the sources
     **/
    public Vector<Source> getSources() {
	return sources;
    }
    
    /**
     * Get all of the functions in this Program
     * @return A Collection containing the TaskDeclarations in the program
     **/
    public Collection<TaskDeclaration> getFunctions() {
        return this.decls.values();
    }
    
    
    /**
     * Get all of the types in this program
     * @return A Collection containing the TypeDeclarations
     **/
    public Collection getTypes() {
        return this.type_decs.values();
    }
    
    /**
     * Get the main function.  The main function (not to be confused with
     * the kickstart) is the expression whose left-hand side appears in
     * no right hand sides.
     * @return The FlowStatement which is the main expression for this program
     **/
    public FlowStatement getMain() {
        Vector stmts = (Vector) this.exps.clone();
        Iterator it = stmts.iterator();
        while (it.hasNext()) {
            FlowStatement stmt = (FlowStatement)it.next();
            String name = stmt.getAssignee();
            for (int i=0; i < this.exps.size(); i++) {
                FlowStatement temp = this.exps.get(i);
                if (temp.inRightHandSide(name)) {
                    it.remove();
                }
            }
        }
        System.out.println(stmts.toString());
        return (FlowStatement)stmts.get(0);
    }
    
    /**
     * Get a named type
     * @param name The name of the type
     * @return The type, or null if it can't be found.
     **/
    public TypeDeclaration getType(String name) {
        return this.type_decs.get(name);
    }
    
    /**
     * Get a named task
     * @param name The name of the type
     * @return The task, or null if it can't be found.
     **/
    public TaskDeclaration getTask(String name) {
        return this.decls.get(name);
    }
    
    /**
     * Get all expressions
     * @return a Vector of FlowStatements containing all expressions
     **/
    public Vector getExpressions() {
        return this.exps;
    }
    
    /**
     * Get an indexe expression
     * @param ix The index of the expression
     * @return The expression
     **/
    public FlowStatement getExpression(int ix) {
        return this.exps.get(ix);
    }
    
    /**
     * Get a named flow statement
     * @param name The name of the left-hand side of the statement
     * @return The FlowStatement whose left-hand side matches name, or null
     **/
    public FlowStatement getFlow(String name) {
        for (int i=0; i < this.exps.size(); i++) {
            if ((this.exps.get(i)).getAssignee().equals(name))
                return this.exps.get(i);
        }
        return null;
    }
    
    /**
     * Find all instances of SimpleFlowStatements matching the left-hand
     * side of the specified statement, remove them from the program
     * and create a TypedFlowStatement containing them.
     * @param stmt The statement to unify
     * @return The new compound TypedFlowStatement
     **/
    public TypedFlowStatement unifyExpression(FlowStatement stmt) {
        TypedFlowStatement result = new TypedFlowStatement(stmt.getAssignee());
        result.addFlowStatement(stmt);
        Iterator it = this.exps.iterator();
        while (it.hasNext()) {
            FlowStatement fs = (FlowStatement)it.next();
            if (fs != stmt) {
                if (fs.getAssignee().equals(result.getAssignee())) {
                    it.remove();
                    result.addFlowStatement(fs);
                }
            }
        }
        return result;
    }
    
    /**
     * Find all instances of typed SimpleFlowStatements and merge them into
     * compound TypedFlowStatements
     * @see TypedFlowStatement
     **/
    public void unifyExpressions() {
        for (int i=0; i < this.exps.size(); i++) {
            if (this.exps.get(i) instanceof SimpleFlowStatement) {
                SimpleFlowStatement fs = (SimpleFlowStatement)this.exps.get(i);
                if (fs.getTypes() != null) {
                    TypedFlowStatement tfs = unifyExpression(fs);
		    this.exps.set(i, tfs);
                }
            }
        }
    }
    
    /**
     * Verify an expression to insure that input types match output types
     * @param stmt The expression to verify
     * @return true if the expression is valid, false otherwise
     **/
    public boolean verifyExpression(SimpleFlowStatement stmt) {
        Vector types = stmt.getTypes();
        Vector args  = stmt.getArguments();
        String left = stmt.getAssignee();

        TaskDeclaration td = this.decls.get(left);

        if (td == null)
		{
            System.err.println("Error, "+left+" is undefined.");
			return false;
        }

        if (types != null) {
            if (types.size()  != td.getInputs().size()) {
                System.err.println("Type specification for "+
                        stmt.getAssignee()+
                        " doesn't match arguments ("+
                        types.size()+":"+td.getInputs().size()+")");
                return false;
            }
            for (int i=0;i<types.size();i++) {
            	String type = (String)types.get(i);
            	if (!type.equals("*") && this.type_decs.get(type)==null) {
            		System.err.println("Type "+type+" is undefined.");
            		return false;
            	}
            }
        }
        
        if (args.size() > 0) {
            String now = (String)args.get(0);
            TaskDeclaration current = this.decls.get(now);
            if (current == null) {
                System.err.println("Error, "+now+" is undefined.");
                return false;
            }
            if (!current.isInMatch(td)) {
                System.err.println("Inputs  of "+now+" don't match inputs of "+
                        left);
                return false;
            }
            int i=1;
            while (i<args.size()) {
                String next = (String)args.get(i++);
                TaskDeclaration next_td = this.decls.get(next);
                if (next_td == null) {
                    System.err.println("Error, "+next+" is undefined.");
                    return false;
                }
                if (!current.isOutInMatch(next_td)) {
                    System.err.println("Outputs of "+now+
                            " don't match inputs of "+next+".");
                    return false;
                }
                now = next;
                current = next_td;
            }
            if (!current.isOutMatch(td)) {
                System.err.println("Outputs of "+now+
                        " don't match inputs of "+left+".");
                return false;
            }
        }
        return true;
    }

	/**
	 * Verify all expression to insure that input types match output types
	 * @return true if the expression is valid, false otherwise
	 **/
	public boolean verifySources()
	{
		for (Source src : sources)
		{
			TaskDeclaration tds = getTask(src.getSourceFunction());

			if (tds == null)
			{
				System.err.println("Error: can't find task declaration for source " +
					src.getSourceFunction());
				return false;
			}

			if (tds.getInputs().size() != 0)
			{
				System.err.println("Error: source nodes should not have inputs, yet task " +
					src.getSourceFunction() + " does.");
				return false;
			}

			TaskDeclaration tdt = getTask(src.getTarget());

			if (tdt == null)
			{
				System.err.println("Error: can't find task declaration for source " +
					src.getSourceFunction() + " target " + src.getTarget());
				return false;
			}

			if (!tds.isOutInMatch(tdt))
			{
				System.err.println("Error: source " + src.getSourceFunction() +
					" outputs don't match target " + src.getTarget() + " inputs.");
				return false;
			}
		}
		return true;
	}
    
    /**
     * Verify all expression to insure that input types match output types
     * @return true if the expression is valid, false otherwise
     **/
    public boolean verifyExpressions() {
        for (int i=0; i < this.exps.size(); i++) {
            if (!verifyExpression((SimpleFlowStatement)this.exps.get(i))) {
                return false;
            }
        }
        return true;
    }

    /**
     * Attempt to determine types of impliciitly declared expressions
     * @return true upon success
     **/
	public boolean preprocessExpressions()
	{
		Map<String, Vector<String>> deps = new HashMap<String, Vector<String>>();
		Map<String, Vector<String>> predLinks = new HashMap<String, Vector<String>>();
		Vector<String> undeclaredTasks = new Vector<String>();

		for (int i=0; i < this.exps.size(); i++)
		{
			SimpleFlowStatement stmt = (SimpleFlowStatement)this.exps.get(i);
			String left = stmt.getAssignee();
			TaskDeclaration td = this.decls.get(left);

			if (td == null && !undeclaredTasks.contains(left))
				undeclaredTasks.add(left);
		}

		for (int i=0; i < this.exps.size(); i++)
		{
			SimpleFlowStatement stmt = (SimpleFlowStatement)this.exps.get(i);
			String left = stmt.getAssignee();
			Vector args  = stmt.getArguments();
			Vector olddeps;
			Vector<String> expdeps = new Vector<String>();

			for (int n = 0; n < args.size(); n++)
				expdeps.add((String)args.get(n));

			olddeps = deps.get(left);

			if (olddeps == null || olddeps.size() < expdeps.size())
				deps.put(left, expdeps);
		}

		for (int i=0; i < this.exps.size(); i++)
		{
			SimpleFlowStatement stmt = (SimpleFlowStatement)this.exps.get(i);
			String left = stmt.getAssignee();

			int x, y;
			Vector values;

			Vector<String> link = new Vector<String>(4);
			link.add(null);
			link.add(null);
			link.add(null);
			link.add(null);

			for (x=0; x < this.exps.size(); x++)
			{
				SimpleFlowStatement xstmt = (SimpleFlowStatement)this.exps.get(x);

				values = deps.get(xstmt.getAssignee());

				for (y = 0; y < values.size(); y++)
				{
					if (values.get(y).equals(left))
					{
						if (y > 0)
							link.set(0, values.get(y - 1).toString());

						if (y == 0)
							link.set(1, xstmt.getAssignee());

						if (y < values.size() - 1)
							link.set(2, values.get(y + 1).toString());

						if (y == values.size() - 1)
							link.set(3, xstmt.getAssignee());
					}
				}
			}

			if (link.get(0) != null ||
				link.get(1) != null ||
				link.get(2) != null ||
				link.get(3) != null)
				predLinks.put(left, link);
		}

		boolean didsomething = true;

		while(didsomething)
		{
			didsomething = false;

			for (int i=0; i < undeclaredTasks.size(); i++)
			{
				SimpleFlowStatement stmt = null;
				String task = undeclaredTasks.get(i);
				Vector expdeps;
				int n, lookup;

				expdeps = deps.get(task);

				if (expdeps.size() == 0)
					continue;

				for (n = 0; n < expdeps.size(); n++)
					if (undeclaredTasks.contains(expdeps.get(n)))
						break;

				if (n < expdeps.size())
					continue;

				stmt = null;

				for (lookup=0; lookup < this.exps.size(); lookup++)
				{
					SimpleFlowStatement istmt = (SimpleFlowStatement)this.exps.get(lookup);
					String left = istmt.getAssignee();

					if (left.equals(task) && (stmt == null || stmt.getArguments().size() < istmt.getArguments().size()))
						stmt = istmt;
				}

				if (stmt == null)
					continue;

				didsomething = true;
				undeclaredTasks.remove(task);

        		Vector types = stmt.getTypes();
    		    Vector args  = stmt.getArguments();
		        TaskDeclaration td;

				Vector<Argument> inputs = new Vector<Argument>();
				Vector<Argument> outputs = new Vector<Argument>();

				//System.out.println("Adding " + task);

	            td = this.decls.get((String)args.get(0));
	            if (td == null)
				{
	                System.err.println("Error, "+(String)args.get(0)+" is undefined.");
	                return false;
	            }
				inputs.addAll(td.getInputs());

                td = this.decls.get((String)args.get(args.size() - 1));
                if (td == null)
				{
                    System.err.println("Error, "+(String)args.get(args.size() - 1)+" is undefined.");
                    return false;
                }
				outputs.addAll(td.getOutputs());

				td = new TaskDeclaration(inputs, task, outputs);

				this.decls.put(task, td);

				/*if (td.getInputs().size() == 0)
					System.out.println("Assuming no inputs");

				for (int z = 0; z < td.getInputs().size(); z++)
					System.out.println("Assumed Input: " + td.getInputs().get(z));

				if (td.getOutputs().size() == 0)
					System.out.println("Assuming no outputs");

				for (int z = 0; z < td.getOutputs().size(); z++)
					System.out.println("Assumed Output: " + td.getOutputs().get(z));*/
			}

			if (didsomething)
				continue;

			for (int i=0; i < undeclaredTasks.size(); i++)
			{
				SimpleFlowStatement stmt = null;
				String task = undeclaredTasks.get(i);
				Vector<String> values = predLinks.get(task);
				int n, lookup;

				if (values == null)
					continue;

				if (values.get(0) != null)
				{
					if (undeclaredTasks.contains(values.get(0)))
					{
						//System.out.println("Can't manage " + task + " due to " + values.get(0));
						continue;
					}
					//System.out.println(task + " is happy due to input0 " + values.get(0));
				}
				else if (values.get(1) != null)
				{
					if (undeclaredTasks.contains(values.get(1)))
					{
						//System.out.println("Can't manage " + task + " due to " + values.get(1));
						continue;
					}
					//System.out.println(task + " is happy due to input1 " + values.get(1));
				}

				if (values.get(2) != null)
				{
					if (undeclaredTasks.contains(values.get(2)))
					{
						//System.out.println("Can't manage " + task + " due to " + values.get(2));
						continue;
					}
					//System.out.println(task + " is happy due to output2 " + values.get(2));
				}
				else if (values.get(3) != null)
				{
					if (undeclaredTasks.contains(values.get(3)))
					{
						//System.out.println("Can't manage " + task + " due to " + values.get(3));
						continue;
					}
					//System.out.println(task + " is happy due to output3 " + values.get(3));
				}

				stmt = null;

				for (lookup=0; lookup < this.exps.size(); lookup++)
				{
					SimpleFlowStatement istmt = (SimpleFlowStatement)this.exps.get(lookup);
					String left = istmt.getAssignee();

					if (left.equals(task) && (stmt == null || stmt.getArguments().size() < istmt.getArguments().size()))
						stmt = istmt;
				}

				if (stmt == null)
					continue;

				didsomething = true;
				undeclaredTasks.remove(task);

        		Vector types = stmt.getTypes();
    		    Vector args  = stmt.getArguments();
		        TaskDeclaration td;

				Vector<Argument> inputs = new Vector<Argument>();
				Vector<Argument> outputs = new Vector<Argument>();

				//System.out.println("Adding " + task);

				int x, y;
				String inputNode = null, outputNode = null;

				if (values.get(0) != null)
				{
					inputNode = values.get(0);
		            td = this.decls.get(inputNode);
		            if (td == null)
					{
		                System.err.println("Error, "+inputNode+" is undefined.");
		                return false;
		            }
					inputs.addAll(td.getOutputs());
				}
				else if (values.get(1) != null)
				{
					inputNode = values.get(1);
		            td = this.decls.get(inputNode);
		            if (td == null)
					{
		                System.err.println("Error, "+inputNode+" is undefined.");
		                return false;
		            }
					inputs.addAll(td.getInputs());
				}

				if (values.get(2) != null)
				{
					outputNode = values.get(2);
		            td = this.decls.get(outputNode);
		            if (td == null)
					{
		                System.err.println("Error, "+outputNode+" is undefined.");
		                return false;
		            }
					outputs.addAll(td.getInputs());
				}
				else if (values.get(3) != null)
				{
					outputNode = values.get(3);
		            td = this.decls.get(outputNode);
		            if (td == null)
					{
		                System.err.println("Error, "+outputNode+" is undefined.");
		                return false;
		            }
					outputs.addAll(td.getOutputs());
				}

				td = new TaskDeclaration(inputs, task, outputs);

				this.decls.put(task, td);

				/*if (td.getInputs().size() == 0)
					System.out.println("Assuming no inputs");

				for (int z = 0; z < td.getInputs().size(); z++)
					System.out.println("Assumed Input: " + td.getInputs().get(z));

				if (td.getOutputs().size() == 0)
					System.out.println("Assuming no outputs");

				for (int z = 0; z < td.getOutputs().size(); z++)
					System.out.println("Assumed Output: " + td.getOutputs().get(z));*/
			}

			if (didsomething)
				continue;

			for (int i=0; i < undeclaredTasks.size(); i++)
			{
				SimpleFlowStatement stmt = null;
				String task = undeclaredTasks.get(i);
				Vector<String> values = predLinks.get(task);
				Vector expdeps;
				int n, lookup;

				expdeps = deps.get(task);

				if (expdeps.size() != 0 || values != null)
					continue;

				Vector<Argument> inputs = new Vector<Argument>();
				Vector<Argument> outputs = new Vector<Argument>();

				TaskDeclaration td = new TaskDeclaration(inputs, task, outputs);

				this.decls.put(task, td);

				didsomething = true;
				undeclaredTasks.remove(task);

				System.err.println("Warning: " + task + " is an implicitly declared abstract node, but is unused.");
			}
		}

		if (undeclaredTasks.size() == 0)
			return true;
		else
		{
			System.err.println("Error determining implicit task declaration from SimpleFlowStatements:");

			for (int i = 0; i < undeclaredTasks.size(); i++)
				System.err.println(undeclaredTasks.get(i));

			return false;
		}
	}
    
    /**
     * String representation of this program
     * @return The string representing this program
     **/
    public String toString() {
        return this.decls+"\n::\n"+this.exps;
    }

    public void fixDeadlock(Vector<AtomicDeclaration> list) {
	Lock large = null;
	AtomicDeclaration last = null;
	Hashtable<Lock, Boolean> has = new Hashtable<Lock, Boolean>();
	
	for (AtomicDeclaration ad : list) {
	    Vector<Lock> lks = ad.getLocks();
	    Lock temp = lks.get(0);
	    for (Lock l : lks) {
		if (l.compareTo(temp) > 0) {
		    temp = l;
		}
	    }
	    if (large == null)
		large = temp;
	    else {
		if (large.compareTo(temp) < 0 && has.get(temp)==null) {
		    System.err.println
			("Attempting to fix deadlock by elevating Lock: "+temp+
			 " to node "+last);
		    last.addLock(temp);
		    return;
		}
	    }
	    for (Lock l : lks) {
		if (has.get(l) == null) {
		    has.put(l, Boolean.TRUE);
		}
	    }
	    last = ad;
	}
	System.err.println("Error, no deadlock to fix!");
    }

    public Vector<AtomicDeclaration> checkDeadlock() {
	for (FlowStatement fs : exps) {
	    Vector<Vector<AtomicDeclaration> > lock_list;
	    if (fs instanceof SimpleFlowStatement) {
		lock_list = buildLockList((SimpleFlowStatement)fs);
	    }
	    else {
		lock_list = buildLockList((TypedFlowStatement)fs);
	    }
	    for (Vector<AtomicDeclaration> list : lock_list) {
		Hashtable<Lock, Boolean> has = new Hashtable<Lock, Boolean>();
		if (list.size() == 0)
		    break;
		Lock large = null;
		for (AtomicDeclaration ad : list) {
		    Vector<Lock> lks = ad.getLocks();
		    Lock temp = lks.get(0);
		    for (Lock l : lks) {
			if (l.compareTo(temp) > 0) {
			    temp = l;
			}
		    }
		    if (large == null)
			large = temp;
		    else {
			if (large.compareTo(temp) < 0 && has.get(temp)==null) {
			    return list;
			}
		    }
		     for (Lock l : lks) {
			if (has.get(l) == null) {
			    has.put(l, Boolean.TRUE);
			}
		     }
		}
	    }
	}
	return null;
    }

    public Vector<Vector<AtomicDeclaration> >
	buildLockList(TypedFlowStatement fs) {
	Vector<Vector<AtomicDeclaration> > result = 
	    new Vector<Vector<AtomicDeclaration> >();
	AtomicDeclaration ad = atomic_defs.get(fs.getAssignee());
	if (ad == null) 
	    return result;
	
	Vector<FlowStatement> flows = fs.getFlowStatements();
	for (FlowStatement nfs : flows) {
	    Vector<Vector<AtomicDeclaration> > nv;
	    if (nfs instanceof SimpleFlowStatement) {
		nv = buildLockList((SimpleFlowStatement)nfs);
	    }
	    else {
		nv = buildLockList((TypedFlowStatement)nfs);
	    }
	    
	    for (Vector<AtomicDeclaration> vx : nv) {
		result.add(vx);
	    }
	}
	return result;
    }

    public Vector<Vector<AtomicDeclaration> > 
	buildLockList(SimpleFlowStatement fs) 
    {
	Vector<Vector<AtomicDeclaration> > result = 
	    new Vector<Vector<AtomicDeclaration> >();
	
	Vector<AtomicDeclaration> v = new Vector<AtomicDeclaration>();
	AtomicDeclaration ad = atomic_defs.get(fs.getAssignee());
	if (ad == null) 
	    return result;
	v.add(ad);
	result.add(v);
	Vector<String> args = fs.getArguments();
	for (String arg : args) {
	    FlowStatement nfs = getFlow(arg);
	    if (nfs == null) // Its a concrete statement
		{
		    ad = atomic_defs.get(arg);
		    if (ad != null) {
			for (Vector<AtomicDeclaration> vt : result) {
			    vt.add(ad);
			}
		    }
		}
	    else {
		Vector<Vector<AtomicDeclaration> > nv;
		if (nfs instanceof SimpleFlowStatement) {
		    nv = buildLockList((SimpleFlowStatement)nfs);
		}
		else {
		    nv = buildLockList((TypedFlowStatement)nfs);
		}

		Vector<Vector<AtomicDeclaration> > new_result = 
		    new Vector<Vector<AtomicDeclaration> >();

		for (Vector<AtomicDeclaration> vt : result) {
		    for (Vector<AtomicDeclaration> vx : nv) {
			Vector<AtomicDeclaration> new_vec =
			    new Vector<AtomicDeclaration>();
			for (AtomicDeclaration atom : vt) 
			    new_vec.add(atom);
			for (AtomicDeclaration atom : vx)
			    new_vec.add(atom);
			new_result.add(new_vec);
		    }
		}
		result = new_result;
	    }
	}
	return result;
    }
		    
	

    /**
     * Get a source by name
     * @param name The name
     * @param srcs The list of sources
     **/
    Source getSource(String name) {
	for (Source src : sources) {
	    if (src.getSourceFunction().equals(name))
		return src;
	}
	return null;
    }

        /**
     * Is this task declaration a source?
     * @param task The task
     * @return true if the task declaration is a source
     **/
    boolean isSource(TaskDeclaration td) {
	String name = td.getName();
	for (Source src : sources) {
	    if (src.getSourceFunction().equals(name))
		return true;
	}
	return false;
    }

}
