package edu.umass.cs.flux;

import java.util.*;
import java.io.*;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import jdsl.graph.ref.IncidenceListGraph;
import jdsl.graph.api.*;

/**
 * This Class takes in a ProgramGraph representation of a Flux graph
 * and outputs a CSIM file that can be used to simulate the cunning time.
 * 
 **/
public class SimulatorGenerator 
{
	ProgramGraph programGraph;

	Collection<TaskDeclaration> functions;

	Collection<Source> sources;

	IncidenceListGraph graph;

	Program program;

	Map<String, Integer> funcnums;
	
	//Vector<Vertex> functionVector;
	SimulatorGenerator(Program p, ProgramGraph pg, String root) 
	{
		program = p;
		programGraph = pg;
		functions = programGraph.getProgram().getFunctions();
		sources = programGraph.getSources();
		graph = programGraph.graph;
		makeCSimProgram(root);
	}

	private void makeCSimProgram(String root) 
	{
		try 
		{
			ArrayList locklist = new ArrayList();
			funcnums = new HashMap<String, Integer>();
			VirtualDirectory virtDir = new VirtualDirectory(root);

			PrintWriter out = new PrintWriter(new BufferedWriter(virtDir
					.getWriter("simulator.cpp")));

			// includes here
			out.println("#include \"cpp.h\"");
			out.println("#include <stdio.h>");
			out.println("#include \"flux_csim_includes.h\"");

			// output file pointer
			out.println("FILE *fp;\n");
			out.println("#define NUM_JOBS\t\t100000");
			out.println("#define TEST_TIME\t\t240000\n");
			out.println("#define THREADPOOL_SIZE\t50");
			out.println("#define NUM_PROCS\t\t1");

			out.println("\nfacility_ms *processor;");
			out.println("\n#ifdef THREADPOOL_SIZE\n" +
						"buffer *thread_pool;\n" +
						"#endif");

			// now we do completions counters
			out.println("\n//completions counters");
			for (Source s : sources) {
				Vertex vertex = (Vertex) programGraph.vertex_map.get(s
						.getSourceFunction());
				//<node_name>_completion_counter = 0;
				String currCompletioinCounter = s.getSourceFunction() + "_completion_counter";
				out.println("unsigned int " + currCompletioinCounter + " = 0;");
			}

			out.println("\n//completion events");
			for (Source s : sources) {
				Vertex vertex = (Vertex) programGraph.vertex_map.get(s
						.getSourceFunction());
				//<node_name>_done_event;
				String currCompletioinCounter = s.getSourceFunction() + "_done_event";
				out.println("event *" + currCompletioinCounter + ";");
			}
			
			{
				Collection<AtomicDeclaration> decs = program.getAtomicDeclarations();

				if (decs.size() > 0)
				{
					out.println("\n// global locks");

					for (AtomicDeclaration ad : decs)
					{
						Vector<Lock> lks = ad.getLocks();

						for (Lock l : lks)
						{
							if(l.isProgram() && !locklist.contains(l.getName() + "_program_lock"))
							{
								locklist.add(l.getName() + "_program_lock");
								out.println("facility *" + l.getName() + "_program_lock;");
							}
							
						}
					}
				}
			}

			out.println("\n// Counters");
			for (TaskDeclaration td : functions) 
			{
				out.println("int num_" + td.getName() + " = 0;");
				if (td.getErrorHandler() != null)
				{
					out.println("int num_" + td.getErrorHandler() + " = 0;");
				}
			}

			out.println("\n// Task decls");
			for (Source s : sources) 
			{
				Vertex vertex = (Vertex) programGraph.vertex_map.get(s.getSourceFunction());
				printForwardDecls(out, vertex);
			}

			// Used for task decls
			funcnums.clear();

			out.println("\n// source wrapper functions");
			for (Source s : sources) {
				Vertex vertex = (Vertex) programGraph.vertex_map.get(s
						.getSourceFunction());
				out.println("void " + s.getSourceFunction() + "_wrapper();");
			}
			
			out.println("\n// internal functions\n" + "void init();");
			out.println("extern \"C\" void sim();\n");

			// output the main function 
			printMain(out);

			// output the init() function here
			printInit(out, locklist);
			
			// output the "sim" function
			printSim(out);

			for (Source s : sources) {
				Vertex vertex = (Vertex) programGraph.vertex_map.get(s
						.getSourceFunction());
				printWrapper(out, s);
			}

			for (Source s : sources) 
			{
				Vertex vertex = (Vertex) programGraph.vertex_map.get(s.getSourceFunction());
				printFunction(out, vertex);
			}

			out.flush();
			out.close();
		} 
		catch (Exception e) 
		{
			System.out.println(e);
		}
	}

	private void printSim(PrintWriter out)
	{
		out.println("extern \"C\" void sim(){");
		
		out.println("	create(\"sim\");\n\n" +
					"#ifdef THREADPOOL_SIZE\n" +
					"	thread_pool->put(THREADPOOL_SIZE);;\n" +
					"#endif\n");

		for (Source s : sources) {
			Vertex vertex = (Vertex) programGraph.vertex_map.get(s
					.getSourceFunction());
			out.println("	" + s.getSourceFunction() + "_wrapper();");
		}
		
		out.println("}\n");
	}
	
	
	/*
	 * The init() function sets up all of the facilities (shared resources). 
	 * Currently, "processor" is a simple "facility" object, but if we want to
	 * simmulate a multiple cpu machine we have to use "facility_ms".
	 * */
	private void printInit(PrintWriter out, ArrayList locks) 
	{
		int i;

		out.println("void init(){");
		out.println("	processor = new facility_ms(\"processor\", NUM_PROCS);\n");

		for (i = 0; i < locks.size(); i++)
			 out.println("	" + locks.get(i) +
						 " = new facility(\""+
						 locks.get(i) + "\");");

		for (Source s : sources) {
			Vertex vertex = (Vertex) programGraph.vertex_map.get(s
					.getSourceFunction());
			out.println("	" + s.getSourceFunction() + "_done_event = new event(\"" +
				s.getSourceFunction() + "_done_event\");");
		}

		out.println("\n#ifdef THREADPOOL_SIZE\n" +
					"	thread_pool = new buffer(\"thread pool\", THREADPOOL_SIZE);\n" +
					"#endif");

		out.println("}\n");
	}

	private void printMain(PrintWriter out) 
	{
		out.println("int main()\n" +
					"{\n" +
					"	max_processes(1000*1000);\n" +
					"\n" +
					"	init();\n" +
					"	fp = fopen(\"results.out\", \"w\");\n" +
					"	fprintf(fp, \"Start Trial: %f\", simtime());\n" +
					"	sim();");

		for (Source s : sources) {
			Vertex vertex = (Vertex) programGraph.vertex_map.get(s
					.getSourceFunction());
			//<node_name>_done_event;
			String currCompletioinCounter = s.getSourceFunction() + "_done_event";
			out.println("	" +currCompletioinCounter + "->wait();");
		}

		out.println("	fprintf(fp, \"End Trial: %f\", simtime());\n");

		for (TaskDeclaration td : functions) 
		{
			out.println("	printf(\"num_" + td.getName() + " %d\\n\", num_" +
						td.getName() + ");");
		}

		out.println("");

		for (Source s : sources) {
			Vertex vertex = (Vertex) programGraph.vertex_map.get(s
					.getSourceFunction());
			out.println("	printf(\"" + s.getSourceFunction() +
				"_completion_counter: %d\\n\", " +
				s.getSourceFunction() + "_completion_counter);");

			out.println("	printf(\"" + s.getSourceFunction() +
				" completions /s: %f\\n\", " +
				"(double) (1000*" +
				s.getSourceFunction() + "_completion_counter / simtime()));");
		}

		out.println("\n	fclose(fp);\n" +
					"	return 0;\n" +
					"}\n");
	}

	private void printWrapper(PrintWriter out, Source s)
	{
		out.println("void " + s.getSourceFunction() + "_wrapper()\n" +
					"{\n" +
					"	create(\"" + s.getSourceFunction() + "_wrapper()\");\n\n" +
					"	while(simtime() < TEST_TIME)\n" +
					"	{\n" +
					"		" + s.getSourceFunction() + "();\n" +
					"		hold(exponential(IA_TIME_" +
						s.getSourceFunction().toUpperCase() +
						"));\n" +
					"	}\n" +
					"	" + s.getSourceFunction() + "_done_event->set();\n" +
					"}\n\n");
	}

	private void printForwardDecls(PrintWriter out, Vertex vertex) {
		GraphNode gNode = (GraphNode) vertex.element();
		String name;

		name = gNode.toString();

		if (funcnums.containsKey(name))
		{
			Integer i = funcnums.get(name);
			funcnums.put(name, i + 1);
			name = name + i;
		}
		else
		{
			funcnums.put(name, 1);
		}

		out.println("void " + name + "();");

		EdgeIterator edgeIterator = this.graph.incidentEdges(vertex,
				EdgeDirection.OUT);
		while (edgeIterator.hasNext()) {
			Edge edge = edgeIterator.nextEdge();
			Vertex destination = this.graph.destination(edge);

			GraphNode graphNode = (GraphNode) destination.element();
			int nodeType = graphNode.getNodeType();
			if (nodeType != GraphNode.ENTRY && nodeType != GraphNode.EXIT
					&& nodeType != GraphNode.ERROR) {
				printForwardDecls(out, destination);

				//vertexVector.add(destination);
			}
		}
		//printFunction(out, gNode.toString(), vertexVector, gNode.toString());
	}

	private void printFunction(PrintWriter out, Vertex vertex) {
		GraphNode gNode = (GraphNode) vertex.element();
		Vector<Vertex> vertexVector = new Vector<Vertex>();

		EdgeIterator edgeIterator = this.graph.incidentEdges(vertex,
				EdgeDirection.OUT);
		while (edgeIterator.hasNext()) {
			Edge edge = edgeIterator.nextEdge();
			Vertex destination = this.graph.destination(edge);
			GraphNode graphNode = (GraphNode) destination.element();
			int nodeType = graphNode.getNodeType();
			if (nodeType != GraphNode.ENTRY && nodeType != GraphNode.EXIT
					&& nodeType != GraphNode.ERROR) {
				vertexVector.add(destination);
			}
		}
		printFunction(out, gNode.toString(), vertexVector, gNode.toString());
	}

	private void printFunction(PrintWriter out, String fnName,
			Vector<Vertex> vertexVector, String currSrcFn) {
		String fnUName;
		boolean isSource = false;

		if (funcnums.containsKey(fnName))
		{
			Integer i = funcnums.get(fnName);
			funcnums.put(fnName, i + 1);
			fnUName = fnName + i;
		}
		else
		{
			funcnums.put(fnName, 1);
			fnUName = fnName;
		}

		out.println("void " + fnUName + "(){\n" +
					"	num_" + fnName + "++;");
		
		String programLock = null;

		if (fnName == currSrcFn)
		{
			out.println("	create(\"" + fnName + "\");");
			isSource = true;
		}

		 out.println("");

		// check if there are any locks
		AtomicDeclaration a = program.getAtomicDeclaration(fnName);
		if (a != null)
		{
			Vector<Lock> locks = a.getLocks();
			for(Lock l: locks)
			{
				if (l.isProgram())
				{
					programLock = "	" + l.getName() + "_program_lock";
					out.println(programLock + "->reserve();");
				}
			}
		}
		String str_cpu = "CMP_TIME_CPU_" + fnName.toUpperCase();
		String str_clock = "CMP_TIME_CLOCK_" + fnName.toUpperCase();
		
		out.println("	processor->reserve();");
		out.println("	hold(exponential(" + str_cpu + "));");
		out.println("	processor->release();");
		
		out.println("	hold(exponential(" + str_clock +  " - " + str_cpu + "));");

		if (a != null)
		{
			Vector<Lock> locks = a.getLocks();
			for(Lock l: locks)
			{
				if (l.isProgram())
				{
					programLock = "	" + l.getName() + "_program_lock";
					out.println(programLock + "->release();");
				}
			}
		}

		out.println("");
		
		if (vertexVector.size() == 0) 
		{
			/*
			 * Completion Counter
			 * int <source_node_name>_completion_counter = 0;
			 * */
			//TaskDeclaration td =
			//graph.
			Vertex v = programGraph.getVertex(fnName);
			//TaskDeclaration task = (TaskDeclaration)v.element();
			GraphNode g = (GraphNode)v.element();
			if (g.getNodeType() == GraphNode.ERROR_HANDLER) 
			{
				out.println("	// This is an error node");
			}
			else
			{
			out.println("	" + currSrcFn + "_completion_counter++;");
			}
			//out.println("//this is a leaf node!!!");
			out.println("	return;");
			out.println("}\n");
		} 
		// if this node can only go to one other node (besides erroring out)
		else if (vertexVector.size() == 1) 
		{
			String funcname;

			Vertex vertex = vertexVector.elementAt(0);
			GraphNode gNode = (GraphNode) vertex.element();
			out.println("	double ret = uniform(0.0, 1.0);\n");
			String prob = "PROB_" + fnName.toUpperCase() + "_TO_" + gNode.toString().toUpperCase();
			out.println("	if(ret <= " + prob + "){");

			if (funcnums.containsKey(gNode.toString()))
			{
				Integer uid = funcnums.get(gNode.toString());
				funcname = gNode.toString() + uid;
			}
			else
				funcname = gNode.toString();

			if (isSource)
				out.println("#ifdef THREADPOOL_SIZE\n" +
							"		thread_pool->get(1);\n" +
							"#endif");

			out.println("		" + funcname + "();");

			if (isSource)
				out.println("\n#ifdef THREADPOOL_SIZE\n" +
							"		thread_pool->put(1);\n" +
							"#endif");

			out.println("		return;");
			out.println("	}\n");
			out.println("	return;");
			out.println("}\n");

			
			// now we generate the children
			Vector<Vertex> vVector = new Vector<Vertex>();
			EdgeIterator edgeIterator = this.graph.incidentEdges(vertex, EdgeDirection.OUT);
			while (edgeIterator.hasNext())
			{
				Edge edge = edgeIterator.nextEdge();
				Vertex destination = this.graph.destination(edge);
				GraphNode graphNode = (GraphNode) destination.element();
				int nodeType = graphNode.getNodeType();
				if (nodeType != GraphNode.ENTRY && nodeType != GraphNode.EXIT && nodeType != GraphNode.ERROR) 
				{
					vVector.add(destination);
				}
			}
			GraphNode node = (GraphNode) vertex.element();
			printFunction(out, node.toString(), vVector, currSrcFn);

		} 
		else 
		{
			String funcname;

			out.println("	double ret = uniform(0.0, 1.0);");
			out.println("	double curr_prob = 0.0;");
			out.println("	double lower_prob = 0.0;");
			out.println("	double higher_prob = 0.0;\n");

			Vector<String> vectorString = new Vector<String>();

			for (int i = 0; i < vertexVector.size(); i++) {
				Vertex vertex = vertexVector.elementAt(i);
				GraphNode graphNode = (GraphNode) vertex.element();

				if (funcnums.containsKey(graphNode.toString()))
				{
					Integer uid = funcnums.get(graphNode.toString());
					funcname = graphNode.toString() + uid;
				}
				else
					funcname = graphNode.toString();

				out.println("	lower_prob = higher_prob;");
				String prob = "PROB_" + fnName.toUpperCase() + "_TO_"
						+ graphNode.toString().toUpperCase();
				out.println("	higher_prob += " + prob + ";");

				out.println("	if (ret > lower_prob  && ret <= higher_prob){");

				if (isSource)
					out.println("#ifdef THREADPOOL_SIZE\n" +
								"		thread_pool->get(1);\n" +
								"#endif");

				out.println("		" + funcname + "();");

				if (isSource)
					out.println("\n#ifdef THREADPOOL_SIZE\n" +
								"		thread_pool->put(1);\n" +
								"#endif");

				out.println("		return;");
				out.println("	}");
			}
			out.println("}\n\n");

			// now we generate the children
			for (int i = 0; i < vertexVector.size(); i++) {
				Vector<Vertex> vVector = new Vector<Vertex>();
				Vertex vertex = vertexVector.elementAt(i);
				EdgeIterator edgeIterator = this.graph.incidentEdges(vertex,
						EdgeDirection.OUT);
				while (edgeIterator.hasNext()) {
					Edge edge = edgeIterator.nextEdge();
					Vertex destination = this.graph.destination(edge);
					GraphNode graphNode = (GraphNode) destination.element();
					int nodeType = graphNode.getNodeType();
					if (nodeType != GraphNode.ENTRY
							&& nodeType != GraphNode.EXIT
							&& nodeType != GraphNode.ERROR) {
						vVector.add(destination);
					}
				}
				GraphNode gNode = (GraphNode) vertex.element();
				printFunction(out, gNode.toString(), vVector, currSrcFn);
			}
		}
	}

	public static void main(String[] args) {

	}
}



/*
 * Here is how different variables are named:
 * 
 * Probabilities:
 * PROB_<src>_TO_<dest>
 * 
 * Inter-Arrival Time
 * IA_TIME_<node-name>
 * 
 * CPU Completion Time
 * CMP_TIME_CPU_<node name>
 * 
 * Wall-Clock Completion Time
 * CMP_TIME_CLOCK_<node name>
 * 
 * Program Locks
 * facility *<node_name>_program_lock;
 * 
 * Completion Counter
 * int <source_node_name>_completion_counter = 0;
 * 
 * */
